#  Copyright (C)  2021 Rage Uday Kiran
#
#      This program is free software: you can redistribute it and/or modify
#      it under the terms of the GNU General Public License as published by
#      the Free Software Foundation, either version 3 of the License, or
#      (at your option) any later version.
#
#      This program is distributed in the hope that it will be useful,
#      but WITHOUT ANY WARRANTY; without even the implied warranty of
#      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#      GNU General Public License for more details.
#
#      You should have received a copy of the GNU General Public License
#      along with this program.  If not, see <https://www.gnu.org/licenses/>.

import numpy as _np
from PAMI.faultTolerantFrequentPattern.basic import abstract as _ab

class VBFTMine(_ab._faultTolerantFrequentPatterns):
    """
        VBFTMine is one of the fundamental algorithm to discover fault tolerant frequent patterns in a uncertain transactional database based on
        bitset representation.
        This program employs apriori property (or downward closure property) to  reduce the search space effectively.

        Reference:
        ----------
            Koh, JL., Yo, PW. (2005). An Efficient Approach for Mining Fault-Tolerant Frequent Patterns Based on Bit Vector Representations.
            In: Zhou, L., Ooi, B.C., Meng, X. (eds) Database Systems for Advanced Applications. DASFAA 2005. Lecture Notes in Computer Science,
            vol 3453. Springer, Berlin, Heidelberg. https://doi.org/10.1007/11408079_51


        Attributes:
        ----------
                iFile : str
                    Input file name or path of the input file
                oFile : str
                    Name of the output file or the path of output file
                minSup: float or int or str
                    The user can specify minSup either in count or proportion of database size.
                    If the program detects the data type of minSup is integer, then it treats minSup is expressed in count.
                    Otherwise, it will be treated as float.
                    Example: minSup=10 will be treated as integer, while minSup=10.0 will be treated as float
                sep : str
                    This variable is used to distinguish items from one another in a transaction. The default seperator is tab space or \t.
                    However, the users can override their default separator.
                startTime:float
                    To record the start time of the mining process
                endTime:float
                    To record the completion time of the mining process
                finalPatterns: dict
                    Storing the complete set of patterns in a dictionary variable
                memoryUSS : float
                    To store the total amount of USS memory consumed by the program
                memoryRSS : float
                    To store the total amount of RSS memory consumed by the program
                Database : list
                    To store the transactions of a database in list

            Methods:
            -------
                startMine()
                    Mining process will start from here
                getPatterns()
                    Complete set of patterns will be retrieved with this function
                save(oFile)
                    Complete set of frequent patterns will be loaded in to a output file
                getPatternsAsDataFrame()
                    Complete set of frequent patterns will be loaded in to a dataframe
                getMemoryUSS()
                    Total amount of USS memory consumed by the mining process will be retrieved from this function
                getMemoryRSS()
                    Total amount of RSS memory consumed by the mining process will be retrieved from this function
                getRuntime()
                    Total amount of runtime taken by the mining process will be retrieved from this function
                candidateToFrequent(candidateList)
                    Generates frequent patterns from the candidate patterns
                frequentToCandidate(frequentList, length)
                    Generates candidate patterns from the frequent patterns


            Executing the code on terminal:
            -------------------------------

                Format:
                ------
                    python3 VBFTMine.py <inputFile> <outputFile> <minSup> <itemSup> <minLength> <faultTolerance>

                Examples:
                ---------
                    python3 VBFTMine.py sampleDB.txt patterns.txt 10.0 3.0 3 1  (minSup will be considered in times of minSup and count of database transactions)

                    python3 VBFTMine.py sampleDB.txt patterns.txt 10  3 2 1    (minSup will be considered in support count or frequency)


            Sample run of the importing code:
            ---------------------------------

                import PAMI.faultTolerantFrequentPattern.basic.VBFTMine as alg

                obj = alg.VBFTMine(iFile, minSup, itemSup, minLength, faultTolerance)

                obj.startMine()

                faultTolerantFrequentPatterns = obj.getPatterns()

                print("Total number of Fault Tolerant Frequent Patterns:", len(faultTolerantFrequentPatterns))

                obj.save(oFile)

                Df = obj.getPatternInDataFrame()

                print("Total Memory in USS:", obj.getMemoryUSS())

                print("Total Memory in RSS", obj.getMemoryRSS())

                print("Total ExecutionTime in seconds:", obj.getRuntime())

            Credits:
            --------
                The complete program was written by P.Likhitha  under the supervision of Professor Rage Uday Kiran.

        """

    _minSup = float()
    _itemSup = float()
    _minLength = int()
    _faultTolerance = int()
    _startTime = float()
    _endTime = float()
    _finalPatterns = {}
    _iFile = " "
    _oFile = " "
    _sep = " "
    _plist = []
    _memoryUSS = float()
    _memoryRSS = float()
    _Database = []
    _mapSupport = {}

    def _creatingItemSets(self):
        """
            Storing the complete transactions of the database/input file in a database variable


        """
        self._Database = []
        if isinstance(self._iFile, _ab._pd.DataFrame):
            temp = []
            if self._iFile.empty:
                print("its empty..")
            i = self._iFile.columns.values.tolist()
            if 'Transactions' in i:
                temp = self._iFile['Transactions'].tolist()

            for k in temp:
                self._Database.append(set(k))
        if isinstance(self._iFile, str):
            if _ab._validators.url(self._iFile):
                data = _ab._urlopen(self._iFile)
                for line in data:
                    line.strip()
                    line = line.decode("utf-8")
                    temp = [i.rstrip() for i in line.split(self._sep)]
                    temp = [x for x in temp if x]
                    self._Database.append(set(temp))
            else:
                try:
                    with open(self._iFile, 'r', encoding='utf-8') as f:
                        for line in f:
                            line.strip()
                            temp = [i.rstrip() for i in line.split(self._sep)]
                            temp = [x for x in temp if x]
                            for i in temp:
                                if i not in self._plist:
                                    self._plist.append(i)
                            self._Database.append(set(temp))
                except IOError:
                    print("File Not Found")
                    quit()

    def _convert(self, value):
        """
        To convert the user specified minSup value

        :param value: user specified minSup value

        :return: converted type
        """
        if type(value) is int:
            value = int(value)
        if type(value) is float:
            value = (len(self._Database) * value)
        if type(value) is str:
            if '.' in value:
                value = float(value)
                value = (len(self._Database) * value)
            else:
                value = int(value)
        return value

    def _Count(self, tids):
        count = 0
        for i in tids:
            if i == 1:
                count += 1
        return count

    def _save(self, prefix, suffix, tidsetx):
        if (prefix == None):
            prefix = suffix
        else:
            prefix = prefix + suffix
        prefix = list(set(prefix))
        prefix.sort()
        val = self._Count(tidsetx)
        self._finalPatterns[tuple(prefix)] = val

    def _processEquivalenceClass(self, prefix, itemsets, tidsets):
        if (len(itemsets) == 1):
            i = itemsets[0]
            tidi = tidsets[0]
            self._save(prefix, [i], tidi)
            return
        for i in range(len(itemsets)):
            itemx = itemsets[i]
            if (itemx == None):
                continue
            tidsetx = tidsets[i]
            classItemsets = []
            classtidsets = []
            itemsetx = [itemx]
            for j in range(i + 1, len(itemsets)):
                itemj = itemsets[j]
                tidsetj = tidsets[j]
                y = list(_np.array(tidsetx) & _np.array(tidsetj))
                total = self._Count(y)
                if total >= self._itemSup:
                    classItemsets.append(itemj)
                    classtidsets.append(y)
            if (len(classItemsets) > 0):
                newprefix = list(set(itemsetx)) + prefix
                self._processEquivalenceClass(newprefix, classItemsets, classtidsets)
            self._save(prefix, list(set(itemsetx)), tidsetx)

    def _oneLengthFrequentItems(self):
        """ To calculate the one Length items"""
        Vector = {}
        items = []
        for i in self._Database:
            for j in self._plist:
                count = 0
                if j in i:
                    count = 1
                if j in Vector:
                    Vector[j].append(count)
                else:
                    Vector[j] = [count]
        for x, y in Vector.items():
            v = self._Count(y)
            if v >= self._itemSup:
                items.append(x)
        return Vector, items

    def startMine(self):
        """
            Frequent pattern mining process will start from here
        """
        self._Database = []
        self._startTime = _ab._time.time()
        self._creatingItemSets()
        self._minSup = self._convert(self._minSup)
        self._itemSup = self._convert(self._itemSup)
        self._minLength = int(self._minLength)
        self._faultTolerance = int(self._faultTolerance)
        Vector, plist = self._oneLengthFrequentItems()
        for i in range(len(self._plist)):
            itemx = plist[i]
            tidsetx = Vector[itemx]
            itemsetx = [itemx]
            itemsets = []
            tidsets = []
            for j in range(i + 1, len(plist)):
                itemj = plist[j]
                tidsetj = Vector[itemj]
                y1 = list(_np.array(tidsetx) & _np.array(tidsetj))
                total = self._Count(y1)
                if total >= self._itemSup:
                    itemsets.append(itemj)
                    tidsets.append(y1)
            if (len(itemsets) > 0):
                self._processEquivalenceClass(itemsetx, itemsets, tidsets)
            self._save(None, itemsetx, tidsetx)
        self._endTime = _ab._time.time()
        process = _ab._psutil.Process(_ab._os.getpid())
        self._memoryUSS = float()
        self._memoryRSS = float()
        self._memoryUSS = process.memory_full_info().uss
        self._memoryRSS = process.memory_info().rss
        print("Frequent patterns were generated successfully using FTApriori algorithm ")

    def getMemoryUSS(self):
        """Total amount of USS memory consumed by the mining process will be retrieved from this function

        :return: returning USS memory consumed by the mining process

        :rtype: float
        """

        return self._memoryUSS

    def getMemoryRSS(self):
        """Total amount of RSS memory consumed by the mining process will be retrieved from this function

        :return: returning RSS memory consumed by the mining process

        :rtype: float
        """

        return self._memoryRSS

    def getRuntime(self):
        """Calculating the total amount of runtime taken by the mining process

        :return: returning total amount of runtime taken by the mining process

        :rtype: float
        """

        return self._endTime - self._startTime

    def getPatternsAsDataFrame(self):
        """Storing final frequent patterns in a dataframe

        :return: returning frequent patterns in a dataframe

        :rtype: pd.DataFrame
        """

        dataFrame = {}
        data = []
        for a, b in self._finalPatterns.items():
            s = str()
            for i in a:
                s = s + i + ' '
            data.append([s, b])
            dataFrame = _ab._pd.DataFrame(data, columns=['Patterns', 'Support'])
        # dataFrame = dataFrame.replace(r'\r+|\n+|\t+',' ', regex=True)
        return dataFrame

    def save(self, outFile):
        """Complete set of frequent patterns will be loaded in to a output file

        :param outFile: name of the output file

        :type outFile: file
        """
        self._oFile = outFile
        writer = open(self._oFile, 'w+')
        for x, y in self._finalPatterns.items():
            s = str()
            for i in x:
                s = s + i + '\t'
            s1 = s.strip() + ":" + str(y)
            writer.write("%s \n" % s1)

    def getPatterns(self):
        """ Function to send the set of frequent patterns after completion of the mining process

        :return: returning frequent patterns

        :rtype: dict
        """
        return self._finalPatterns

    def printResults(self):
        print("Total number of Frequent Patterns:", len(self.getPatterns()))
        print("Total Memory in USS:", self.getMemoryUSS())
        print("Total Memory in RSS", self.getMemoryRSS())
        print("Total ExecutionTime in ms:", self.getRuntime())


if __name__ == "__main__":
    _ap = str()
    if len(_ab._sys.argv) == 7 or len(_ab._sys.argv) == 8:
        if len(_ab._sys.argv) == 8:
            _ap = VBFTMine(_ab._sys.argv[1], _ab._sys.argv[3],  _ab._sys.argv[4],
                            _ab._sys.argv[5], _ab._sys.argv[6], _ab._sys.argv[7],)
        if len(_ab._sys.argv) == 7:
            _ap = VBFTMine(_ab._sys.argv[1], _ab._sys.argv[3], _ab._sys.argv[4], _ab._sys.argv[5], _ab._sys.argv[6])
        _ap.startMine()
        print("Total number of Frequent Patterns:", len(_ap.getPatterns()))
        _ap.save(_ab._sys.argv[2])
        print("Total Memory in USS:", _ap.getMemoryUSS())
        print("Total Memory in RSS", _ap.getMemoryRSS())
        print("Total ExecutionTime in ms:", _ap.getRuntime())
    else:
        _ap = VBFTMine('/Users/Likhitha/Downloads/fault/sample.txt', 6, 5, 3, 2, ' ')
        _ap.startMine()
        _ap.printResults()
        print(_ap.getPatternsAsDataFrame())
        print("Error! The number of input parameters do not match the total number of parameters provided")



import numpy as np
from PAMI.faultTolerantFrequentPattern.maximal import abstract as _ab

class maxFTP(_ab._faultTolerantFrequentPatterns):
    """
        FT-Apriori is one of the fundamental algorithm to discover fault tolerant frequent patterns in a transactional database.
        This program employs apriori property (or downward closure property) to  reduce the search space effectively.

        Reference:
        ----------
            Pei, Jian & Tung, Anthony & Han, Jiawei. (2001). Fault-Tolerant Frequent Pattern Mining: Problems and Challenges.


        Attributes:
        ----------
            iFile : str
                Input file name or path of the input file
            oFile : str
                Name of the output file or the path of output file
            minSup: float or int or str
                The user can specify minSup either in count or proportion of database size.
                If the program detects the data type of minSup is integer, then it treats minSup is expressed in count.
                Otherwise, it will be treated as float.
                Example: minSup=10 will be treated as integer, while minSup=10.0 will be treated as float
            sep : str
                This variable is used to distinguish items from one another in a transaction. The default seperator is tab space or \t.
                    However, the users can override their default separator.
            startTime:float
                To record the start time of the mining process
            endTime:float
                To record the completion time of the mining process
            finalPatterns: dict
                Storing the complete set of patterns in a dictionary variable
            memoryUSS : float
                To store the total amount of USS memory consumed by the program
            memoryRSS : float
                To store the total amount of RSS memory consumed by the program
            Database : list
                To store the transactions of a database in list

        Methods:
        -------
            startMine()
                Mining process will start from here
            getPatterns()
                Complete set of patterns will be retrieved with this function
            save(oFile)
                Complete set of frequent patterns will be loaded in to a output file
            getPatternsAsDataFrame()
                Complete set of frequent patterns will be loaded in to a dataframe
            getMemoryUSS()
                Total amount of USS memory consumed by the mining process will be retrieved from this function
            getMemoryRSS()
                Total amount of RSS memory consumed by the mining process will be retrieved from this function
            getRuntime()
                Total amount of runtime taken by the mining process will be retrieved from this function


        Executing the code on terminal:
        -------------------------------

            Format:
            ------
                python3 FTApriori.py <inputFile> <outputFile> <minSup> <itemSup> <minLength> <faultTolerance>

            Examples:
            ---------
                python3 FTApriori.py sampleDB.txt patterns.txt 10.0 3.0 3 1  (minSup will be considered in times of minSup and count of database transactions)

                python3 FTApriori.py sampleDB.txt patterns.txt 10  3 2 1    (minSup will be considered in support count or frequency)


        Sample run of the importing code:
        ---------------------------------

            import PAMI.faultTolerantFrequentPattern.maximal.maxFTP as alg

            obj = alg.maxFTP(iFile, minSup, itemSup, minLength, faultTolerance)

            obj.startMine()

            faultTolerantFrequentPatterns = obj.getPatterns()

            print("Total number of Fault Tolerant Frequent Patterns:", len(faultTolerantFrequentPatterns))

            obj.save(oFile)

            Df = obj.getPatternInDataFrame()

            print("Total Memory in USS:", obj.getMemoryUSS())

            print("Total Memory in RSS", obj.getMemoryRSS())

            print("Total ExecutionTime in seconds:", obj.getRuntime())

        Credits:
        --------
            The complete program was written by P.Likhitha  under the supervision of Professor Rage Uday Kiran.

    """

    _minSup = float()
    _itemSup = float()
    _minLength = int()
    _faultTolerance = int()
    _startTime = float()
    _endTime = float()
    _finalPatterns = {}
    _iFile = " "
    _oFile = " "
    _sep = " "
    _memoryUSS = float()
    _memoryRSS = float()
    _Database = []
    _mapSupport = {}

    def _creatingItemSets(self):
        """
            Storing the complete transactions of the database/input file in a database variable


        """
        self._Database = []
        if isinstance(self._iFile, _ab._pd.DataFrame):
            temp = []
            if self._iFile.empty:
                print("its empty..")
            i = self._iFile.columns.values.tolist()
            if 'Transactions' in i:
                temp = self._iFile['Transactions'].tolist()

            for k in temp:
                self._Database.append(set(k))
        if isinstance(self._iFile, str):
            if _ab._validators.url(self._iFile):
                data = _ab._urlopen(self._iFile)
                for line in data:
                    line.strip()
                    line = line.decode("utf-8")
                    temp = [i.rstrip() for i in line.split(self._sep)]
                    temp = [x for x in temp if x]
                    self._Database.append(set(temp))
            else:
                try:
                    with open(self._iFile, 'r', encoding='utf-8') as f:
                        for line in f:
                            line.strip()
                            temp = [i.rstrip() for i in line.split(self._sep)]
                            temp = [x for x in temp if x]
                            self._Database.append(set(temp))
                except IOError:
                    print("File Not Found")
                    quit()

    def _convert(self, value):
        """
        To convert the user specified minSup value

        :param value: user specified minSup value

        :return: converted type
        """
        if type(value) is int:
            value = int(value)
        if type(value) is float:
            value = (len(self._Database) * value)
        if type(value) is str:
            if '.' in value:
                value = float(value)
                value = (len(self._Database) * value)
            else:
                value = int(value)
        return value

    def _oneLengthItems(self):
        self._mapSupport = {}
        self._bits = {}
        for i in self._Database:
            for j in i:
                if j not in self._mapSupport:
                    self._mapSupport[j] = 1
                else:
                    self._mapSupport[j] += 1
        self._mapSupport = {k: v for k, v in self._mapSupport.items() if v >= self._minSup}
        genList = [k for k, v in sorted(self._mapSupport.items(), key=lambda x: x[1], reverse=True)]
        for i in genList:
            for j in self._Database:
                if i in j:
                    if i not in self._bits:
                        self._bits[i] = np.array(1)
                    else:
                        self._bits[i].append(1)
                else:
                    if i not in self._bits:
                        self._bits[i] = np.array(0)
                    else:
                        self._bits[i].append(0)
        print(self._bits)
        return genList

    def _checkMaximal(self, itemset):
        s = [0] * len(self._Database)
        for i in itemset:
            s = list(np.array(s) & np.array(self._bits[i]))
        print(s)



    def startMine(self):
        self._startTime =  _ab._time.time()
        self._creatingItemSets()
        self._minSup = self._convert(self._minSup)
        self._itemSup = self._convert(self._itemSup)
        self._faultTolerance = int(self._faultTolerance)
        sortedList = self._oneLengthItems()
        self._checkMaximal(sortedList)
        self._maxFTP(sortedList)
        self.__endTime = _ab._time.time()
        self.__memoryUSS = float()
        self.__memoryRSS = float()
        process = _ab._psutil.Process(_ab._os.getpid())
        self.__memoryUSS = process.memory_full_info().uss
        self.__memoryRSS = process.memory_info().rss

    def getMemoryUSS(self):
        """Total amount of USS memory consumed by the mining process will be retrieved from this function

        :return: returning USS memory consumed by the mining process

        :rtype: float
        """

        return self.__memoryUSS

    def getMemoryRSS(self):
        """Total amount of RSS memory consumed by the mining process will be retrieved from this function

        :return: returning RSS memory consumed by the mining process

        :rtype: float
        """

        return self.__memoryRSS

    def getRuntime(self):
        """Calculating the total amount of runtime taken by the mining process


        :return: returning total amount of runtime taken by the mining process

        :rtype: float
        """

        return self._endTime - self._startTime

    def getPatternsAsDataFrame(self):
        """Storing final frequent patterns in a dataframe

        :return: returning frequent patterns in a dataframe

        :rtype: pd.DataFrame
        """

        dataframe = {}
        data = []
        for a, b in self._finalPatterns.items():
            data.append([a.replace('\t', ' '), b])
            dataframe = _ab._pd.DataFrame(data, columns=['Patterns', 'Support'])
        return dataframe

    def save(self, outFile):
        """Complete set of frequent patterns will be loaded in to a output file

        :param outFile: name of the output file

        :type outFile: file
        """
        self._oFile = outFile
        writer = open(self._oFile, 'w+')
        for x, y in self._finalPatterns.items():
            s1 = x.strip() + ":" + str(y)
            writer.write("%s \n" % s1)

    def getPatterns(self):
        """ Function to send the set of frequent patterns after completion of the mining process

        :return: returning frequent patterns

        :rtype: dict
        """
        return self._finalPatterns

    def printResults(self):
        print("Total number of Frequent Patterns:", len(self.getPatterns()))
        print("Total Memory in USS:", self.getMemoryUSS())
        print("Total Memory in RSS", self.getMemoryRSS())
        print("Total ExecutionTime in ms:", self.getRuntime())


if __name__ == "__main__":
    _ap = str()
    if len(_ab._sys.argv) == 7 or len(_ab._sys.argv) == 8:
        if len(_ab._sys.argv) == 8:
            _ap = maxFTP(_ab._sys.argv[1], _ab._sys.argv[3], _ab._sys.argv[4], _ab._sys.argv[5], _ab._sys.argv[6], _ab._sys.argv[7])
        if len(_ab._sys.argv) == 7:
            _ap = maxFTP(_ab._sys.argv[1], _ab._sys.argv[3], _ab._sys.argv[4], _ab._sys.argv[6], _ab._sys.argv[6])
        _ap.startMine()
        print("Total number of maximal Fault-Tolerant Frequent Patterns:", len(_ap.getPatterns()))
        _ap.save(_ab._sys.argv[2])
        print("Total Memory in USS:", _ap.getMemoryUSS())
        print("Total Memory in RSS", _ap.getMemoryRSS())
        print("Total ExecutionTime in ms:", _ap.getRuntime())
    else:
        _ap = maxFTP('sample.txt', 3, 2, 1, ' ')
        _ap.startMine()
        print("Total number of maximal Fault-Tolerant Frequent Patterns:", len(_ap.getPatterns()))
        _ap.save('output.txt')
        print("Total Memory in USS:", _ap.getMemoryUSS())
        print("Total Memory in RSS", _ap.getMemoryRSS())
        print("Total ExecutionTime in ms:", _ap.getRuntime())
        print("Error! The number of input parameters do not match the total number of parameters provided")




from PAMI.sequentialPatternMining.basic import abstract as _ab


class _Sequence:
    def __init__(self, sequence, support):
        """
        Attributes:
            sequence : list
                A input sequence on which we will create the Sequence object
            support : int
                Support of the given sequence
        """
        self.sequence = []
        self.support = support
        self.place_holder = '_'
        for element in sequence:
            self.sequence.append(list(element))

    def add(self, seq):
        """
        Used when projecting the database by extending a sequence.
        """
        if seq.sequence[0][0] == self.place_holder:
            fEle = seq.sequence[0]
            fEle.remove(self.place_holder)
            self.sequence[-1].extend(fEle)
            print(self.sequence[-1])
            self.sequence.extend(seq.sequence[1:])
        else:
            self.sequence.extend(seq.sequence)
            if self.support == None:
                self.support = seq.support
        self.support = min(self.support, seq.support)


class prefixSpan(_ab._sequentialPatterns):
    """
        PrefixSpan is one of the basic algorithm to discover the frequent sequential patterns.

    Reference:
    ---------
        J. Pei, J. Han, B. Mortazavi-Asl, J. Wang, H. Pinto, Q. Chen, U. Dayal,
        M. Hsu: Mining Sequential Patterns by Pattern-Growth: The PrefixSpan Approach.
        IEEE Trans. Knowl. Data Eng. 16(11): 1424-1440 (2004)

    Attribute:
    ---------
        iFile : str
            Input file name or path of the input file
        oFile : str
            Name of the output file or the path of output file
        minSup: float or int or str
            The user can specify minSup either in count or proportion of database size.
            If the program detects the data type of minSup is integer, then it treats minSup is expressed in count.
            Otherwise, it will be treated as float.
            Example: minSup=10 will be treated as integer, while minSup=10.0 will be treated as float
        sep : str
            This variable is used to distinguish items from one another in a transaction. The default seperator is tab space or \t.
            However, the users can override their default separator.
        startTime:float
            To record the start time of the mining process
        endTime:float
            To record the completion time of the mining process
        finalPatterns: dict
            Storing the complete set of patterns in a dictionary variable
        memoryUSS : float
            To store the total amount of USS memory consumed by the program
        memoryRSS : float
            To store the total amount of RSS memory consumed by the program
        Database : list
            To store the transactions of a database in list

    Methods:
    -------
        startMine()
            Start the Mining Process By calling the recursion function
        recursion()
            The main Function to startMine the database level by level
        projectDataBase()
            This function will project the input databse based on the input pattern
        validateLengths()
            Find whether all the transactions in the sequence is less than the maximum pattern length
        getFrequentItems()
            This function will return the frequent items based in the minimum support that is called in the recursion function
        getTime()
            return the time taken for mining
        getMemory()
            return the maximum memory utilized while mining
    """
    _minSup = float()
    _maxPatternLength = int()
    _PLACEHOLDER = '_'
    _startTime = float()
    _endTime = float()
    _finalPatterns = {}
    _iFile = " "
    _oFile = " "
    _sep = " "
    _memoryUSS = float()
    _memoryRSS = float()
    _Database = []

    def _creatingItemSets(self):
        """
            Storing the complete transactions of the database/input file in a database variable


        """
        self._Database = []
        if isinstance(self._iFile, _ab._pd.DataFrame):
            temp = []
            if self._iFile.empty:
                print("its empty..")
            i = self._iFile.columns.values.tolist()
            if 'Transactions' in i:
                temp = self._iFile['Transactions'].tolist()

            for k in temp:
                self._Database.append(set(k))
        if isinstance(self._iFile, str):
            if _ab._validators.url(self._iFile):
                data = _ab._urlopen(self._iFile)
                for line in data:
                    line = line.strip()
                    tempLis = line.split(' -1 ')
                    tempLis = tempLis[:-1]
                    mLis = []
                    for ele in tempLis:
                        mLis.append(list(map(int, ele.split(self._sep))))
                    self._Database.append(mLis)
            else:
                try:
                    with open(self._iFile, 'r', encoding='utf-8') as f:
                        for line in f:
                            line = line.strip()
                            tempLis = line.split(' -1 ')
                            tempLis = tempLis[:-1]
                            mLis = []
                            for ele in tempLis:
                                mLis.append(list(map(int, ele.split(self._sep))))
                            self._Database.append(mLis)
                except IOError:
                    print("File Not Found")
                    quit()

    def _powerset(self, iterable):
        s = list(iterable)
        return list(_ab._chain.from_iterable(_ab._combinations(s, r) for r in range(len(s) + 1)))

    def _recur(self, output, n, subs, lis):
        if n == len(subs):
            temp = str(output).replace('(), ', '')
            temp = temp.replace(', ()', '')
            temp = temp.replace(',)', ')')
            if temp != '(())' or len(temp) <= 2:
                lis.append(temp)
            return
        for x in subs[n]:
            self._recur(output + (x,), n + 1, subs, lis)

    def _isSubSequence1(self, string1, string2, m, n):
        if m == 0:
            return True
        if n == 0:
            return False
        if string1[m - 1] == string2[n - 1]:
            return self._isSubSequence(string1, string2, m - 1, n - 1)
        return self._isSubSequence(string1, string2, m, n - 1)

    def _isSubSequence(self, string1, string2, m, n):
        if m == 0:
            return 1
        if n == 0:
            return 0
        if string1[m - 1] == string2[n - 1]:
            return self._isSubSequence(string1, string2, m - 1, n - 1)
        return self._isSubSequence(string1, string2, m, n - 1)

    def _FPMUtil(self, sequences, outputFile):
        allPossibleSequences = []
        mainDict = {}
        for sequence in sequences:
            subs = []
            for x in sequence:
                subs.append(list(self._powerset(x)))

            i = len(subs)
            lis = []
            output = ()
            self._recur(output, 0, subs, lis)
            allPossibleSequences.extend(lis)

        allPossibleSequences = list(set(allPossibleSequences))

        for seq in allPossibleSequences:
            for row in sequences:
                row = str(row).replace(',)', ')')
                seq1 = seq[1:-1].split('), ')
                row1 = row[1:-1].split('), ')
                seq1 = [x + ')' for x in seq if x[-1] != ')']
                row1 = [x + ')' for x in row if x[-1] != ')']
                if (mainDict.get(seq)):
                    mainDict[seq] += self._isSubSequence(seq, row, len(seq), len(row))
                else:
                    mainDict[seq] = self._isSubSequence(seq, row, len(seq), len(row))
        self._finalPatterns = mainDict

    def _recursion(self, pattern, inp, threshold, maxPatternLength):
        '''The main Function to startMine and project the database at each level
            @ pattern : Pattern to build upon
            @ inp : Input Database at the current level
            @ threshold : Minimum support
            @ maxPattern Length : Maxium Pattern length
        '''
        resultList = []
        if self._validateLengths(pattern, maxPatternLength):
            freqList = self._getFrequentItems(inp, pattern, threshold, maxPatternLength)
            for item in freqList:
                retVal = _Sequence(pattern.sequence, pattern.support)
                retVal.add(item)
                if self._validateLengths(pattern, maxPatternLength):
                    resultList.append(retVal)
                projectedDB = self._projectDatabase(inp, retVal)
                newPatterns = self._recursion(retVal, projectedDB, threshold, maxPatternLength)
                resultList.extend(newPatterns)
        return resultList

    def _validateLengths(self, pattern, maxPatternLength):
        '''
            Find whether all the transactions in the sequence is less than the maximum pattern length
        '''
        for item in pattern.sequence:
            if len(item) > maxPatternLength:
                return False
        return True

    def _projectDatabase(self, S, pattern):
        '''
            This function will project the input databse based on the input pattern
        '''
        projectedDB = []
        lastEle = pattern.sequence[-1]
        last_item = lastEle[-1]
        for s in S:
            projectTempDB = []
            for element in s:
                isPrefix = False
                if self._PLACEHOLDER in element:
                    if last_item in element and len(pattern.sequence[-1]) > 1:
                        isPrefix = True
                else:
                    isPrefix = True
                    for item in lastEle:
                        if item not in element:
                            isPrefix = False
                            break
                if isPrefix:
                    currentIndex = s.index(element)
                    lastIndex = element.index(last_item)
                    if lastIndex == len(element) - 1:
                        projectTempDB = s[currentIndex + 1:]
                    else:
                        projectTempDB = s[currentIndex:]
                        index = element.index(last_item)
                        slicedEle = element[lastIndex:]
                        slicedEle[0] = self._PLACEHOLDER
                        projectTempDB[0] = slicedEle
                    break
            if len(projectTempDB) != 0:
                projectedDB.append(projectTempDB)
        return projectedDB

    '''def _loadDataFromFile(self, filename):
        sequences = []
        with open(filename) as f:
            for x in f:
                tempLis = x.split(' -1 ')
                tempLis = tempLis[:-1]
                mLis = []
                for ele in tempLis:
                    mLis.append(list(map(int, ele.split())))
                sequences.append(mLis)

        self.db = sequences'''

    def _save(self, output, lis):
        with open(output, 'w+') as file:
            i = 0
            for x in lis:
                st = ' -1 '.join(list(map(lambda z: str(z)[1:-1], x.sequence)))
                st += f' : {x.support}'
                st += '\n'
                st = st.replace(', ', '\t')
                file.write(st)

    def _getFrequentItems(self, S, pattern, threshold, maxPatternLength):
        """
            This function will return the frequent items based in the minimum support that is called in the recursion function
            @ pattern : Pattern to check for the frequency
            @ S : Input Database at the current level
            @ threshold : Minimum support
            @ maxPattern Length : Maxium Pattern length
        """
        items = {}
        parentItemDict = {}
        freqList = []
        if S is None or len(S) == 0:
            return []
        if len(pattern.sequence) != 0:
            lastEle = pattern.sequence[-1]
        else:
            lastEle = []
        for s in S:
            isPrefix = True
            for item in lastEle:
                if item not in s[0]:
                    isPrefix = False
                    break
            if isPrefix and len(lastEle) > 0:
                index = s[0].index(lastEle[-1])
                if index < len(s[0]) - 1:
                    for item in s[0][index + 1:]:
                        if item in parentItemDict:
                            parentItemDict[item] += 1
                        else:
                            parentItemDict[item] = 1
            if self._PLACEHOLDER in s[0]:
                for item in s[0][1:]:
                    if item in parentItemDict:
                        parentItemDict[item] += 1
                    else:
                        parentItemDict[item] = 1
                s = s[1:]
            counted = []
            for element in s:
                for item in element:
                    if item not in counted:
                        counted.append(item)
                        if item in items:
                            items[item] += 1
                        else:
                            items[item] = 1

        freqList.extend([_Sequence([[self._PLACEHOLDER, key]], value)
                         for key, value in parentItemDict.items() if value >= threshold])
        freqList.extend([_Sequence([[key]], value)
                         for key, value in items.items() if value >= threshold])

        freqList = [item for item in freqList if self._validateLengths(item, maxPatternLength)]
        return sorted(freqList, key=lambda x: x.support)

    def startMine(self):
        '''startMine and return the frequent sequence patterns.. '''
        self._startTime = _ab._time.time()
        self._creatingItemSets()
        self._minSup = _ab._math.ceil(self._minSup * len(self._Database))
        if self._minSup == 1:
            mainTupe = []
            for x in self._Database:
                temp = []
                for y in x:
                    temp.append(tuple(y))
                temp = tuple(temp)
                mainTupe.append(temp)
            mainTupe = tuple(mainTupe)
            self._FPMUtil(mainTupe, self._oFile)
            #print('oFile: ', self._oFile)
        result = self._recursion(_Sequence([], None),
                                self._Database,
                                _ab._math.ceil(self._minSup * len(self._Database)),
                                self._maxPatternLength)
        self.endTime = _ab._time.time()
        process = _ab._psutil.Process(_ab._os.getpid())
        self._memoryUSS = float()
        self._memoryRSS = float()
        self._memoryUSS = process.memory_full_info().uss
        self._memoryRSS = process.memory_info().rss
        print(result)
        return result

    def getMemoryUSS(self):
        """Total amount of USS memory consumed by the mining process will be retrieved from this function

        :return: returning USS memory consumed by the mining process

        :rtype: float
        """

        return self._memoryUSS

    def getMemoryRSS(self):
        """Total amount of RSS memory consumed by the mining process will be retrieved from this function

        :return: returning RSS memory consumed by the mining process

        :rtype: float
        """

        return self._memoryRSS

    def getRuntime(self):
        """Calculating the total amount of runtime taken by the mining process

        :return: returning total amount of runtime taken by the mining process

        :rtype: float
        """

        return self._endTime - self._startTime

    def getPatternsAsDataFrame(self):
        """Storing final frequent patterns in a dataframe

        :return: returning frequent patterns in a dataframe

        :rtype: pd.DataFrame
        """

        dataFrame = {}
        data = []
        for a, b in self._finalPatterns.items():
            data.append([a.replace('\t', ' '), b])
            dataFrame = _ab._pd.DataFrame(data, columns=['Patterns', 'Support'])
        # dataFrame = dataFrame.replace(r'\r+|\n+|\t+',' ', regex=True)
        return dataFrame

    def save(self, outFile):
        """Complete set of frequent patterns will be loaded in to a output file

        :param outFile: name of the output file

        :type outFile: file
        """
        self._oFile = outFile
        writer = open(self._oFile, 'w+')
        for x, y in self._finalPatterns.items():
            s1 = x.strip() + ":" + str(y)
            writer.write("%s \n" % s1)

    def getPatterns(self):
        """ Function to send the set of frequent patterns after completion of the mining process

        :return: returning frequent patterns

        :rtype: dict
        """
        return self._finalPatterns

    def printResults(self):
        print("Total number of Frequent Sequential Patterns:", len(self.getPatterns()))
        print("Total Memory in USS:", self.getMemoryUSS())
        print("Total Memory in RSS", self.getMemoryRSS())
        print("Total ExecutionTime in ms:", self.getRuntime())

if __name__ == "__main__":

    _ap = str()
    if len(_ab._sys.argv) == 4 or len(_ab._sys.argv) == 5:
        if len(_ab._sys.argv) == 5:
            _ap = prefixSpan(_ab._sys.argv[1], _ab._sys.argv[3], _ab._sys.argv[4])
        if len(_ab._sys.argv) == 4:
            _ap = prefixSpan(_ab._sys.argv[1], _ab._sys.argv[3])
        _ap.startMine()
        print("Total number of Sequential Frequent Patterns:", len(_ap.getPatterns()))
        _ap.save(_ab._sys.argv[2])
        print("Total Memory in USS:", _ap.getMemoryUSS())
        print("Total Memory in RSS", _ap.getMemoryRSS())
        print("Total ExecutionTime in ms:", _ap.getRuntime())
    else:
        _ap = prefixSpan('/Users/Likhitha/Downloads/MAIN_SPMF/InpDataPrefixSpan', 0.6, 100, ' ')
        _ap.startMine()
        print("Total number of Sequential Frequent Patterns:", len(_ap.getPatterns()))
        _ap.save('/Users/Likhitha/Downloads/MAIN_SPMF/output.txt')
        print("Total Memory in USS:", _ap.getMemoryUSS())
        print("Total Memory in RSS", _ap.getMemoryRSS())
        print("Total ExecutionTime in ms:", _ap.getRuntime())
        print("Error! The number of input parameters do not match the total number of parameters provided")



from PAMI.stablePeriodicFrequentPattern.basic import abstract as _ab

class SPPEclat:
    """  Stable periodic pattern mining aims to dicover all interesting patterns in a temporal database using three contraints minimum support,
         maximum period and maximum lability, that have support no less than the user-specified minimum support  constraint and lability no
          greater than maximum lability.

            Reference:
            --------
                Fournier-Viger, P., Yang, P., Lin, J. C.-W., Kiran, U. (2019). Discovering Stable Periodic-Frequent Patterns in Transactional Data. Proc.
                 32nd Intern. Conf. on Industrial, Engineering and Other Applications of Applied Intelligent Systems (IEA AIE 2019), Springer LNAI, pp. 230-244

        Attributes:
        ---------
            iFile : file
                Name of the Input file or path of the input file
            oFile : file
                Name of the output file or path of the output file
            minSup: int or float or str
                The user can specify minSup either in count or proportion of database size.
                If the program detects the data type of minSup is integer, then it treats minSup is expressed in count.
                Otherwise, it will be treated as float.
                Example: minSup=10 will be treated as integer, while minSup=10.0 will be treated as float
            maxPer: int or float or str
                The user can specify maxPer either in count or proportion of database size.
                If the program detects the data type of maxPer is integer, then it treats maxPer is expressed in count.
                Otherwise, it will be treated as float.
                Example: maxPer=10 will be treated as integer, while maxPer=10.0 will be treated as float
            maxLa: int or float or str
                The user can specify maxLa either in count or proportion of database size.
                If the program detects the data type of maxLa is integer, then it treats maxLa is expressed in count.
                Otherwise, it will be treated as float.
                Example: maxLa=10 will be treated as integer, while maxLa=10.0 will be treated as float
            sep : str
                This variable is used to distinguish items from one another in a transaction. The default seperator is tab space or \t.
                However, the users can override their default separator.
            memoryUSS : float
                To store the total amount of USS memory consumed by the program
            memoryRSS : float
                To store the total amount of RSS memory consumed by the program
            startTime:float
                To record the start time of the mining process
            endTime:float
                To record the completion time of the mining process
            Database : list
                To store the transactions of a database in list
            mapSupport : Dictionary
                To maintain the information of item and their frequency
            lno : int
                it represents the total no of transactions
            tree : class
                it represents the Tree class
            itemSetCount : int
                it represents the total no of patterns
            finalPatterns : dict
                it represents to store the patterns
            tidList : dict
                stores the timestamps of an item

        Methods:
        -------
            startMine()
                Mining process will start from here
            getPatterns()
                Complete set of patterns will be retrieved with this function
            save(oFile)
                Complete set of periodic-frequent patterns will be loaded in to a output file
            getPatternsAsDataFrame()
                Complete set of periodic-frequent patterns will be loaded in to a dataframe
            getMemoryUSS()
                Total amount of USS memory consumed by the mining process will be retrieved from this function
            getMemoryRSS()
                Total amount of RSS memory consumed by the mining process will be retrieved from this function
            getRuntime()
                Total amount of runtime taken by the mining process will be retrieved from this function
            creatingItemSets()
                Scan the database and store the items with their timestamps which are periodic frequent
            calculateLa()
                Calculates the support and period for a list of timestamps.
            Generation()
                Used to implement prefix class equivalence method to generate the periodic patterns recursively

            Executing the code on terminal:
            -------
            Format:
            ------
                python3 SPPECLAT.py <inputFile> <outputFile> <minSup> <maxPer> <maxLa>

            Examples:
            --------
                python3 SPPECLAT.py sampleDB.txt patterns.txt 10.0 4.0 2.0   (constraints will be considered in percentage of database transactions)

                python3 SPPECLAT.py sampleDB.txt patterns.txt 10 4 2    (constraint will be considered in support count or frequency)

            Sample run of the imported code:
            --------------

                from PAMI.stablePeriodicFrequentPattern.basic import SPPECLAT as alg

                obj = alg.PFPECLAT("../basic/sampleTDB.txt", 5, 3, 3)

                obj.startMine()

                Patterns = obj.getPatterns()

                print("Total number of Stable Periodic Frequent Patterns:", len(Patterns))

                obj.save("patterns")

                Df = obj.getPatternsAsDataFrame()

                memUSS = obj.getMemoryUSS()

                print("Total Memory in USS:", memUSS)

                memRSS = obj.getMemoryRSS()

                print("Total Memory in RSS", memRSS)

                run = obj.getRuntime()

                print("Total ExecutionTime in seconds:", run)

        Credits:
        -------
            The complete program was written by P.Likhitha  under the supervision of Professor Rage Uday Kiran.\n

            """
    _iFile = " "
    _oFile = " "
    _minSup = str()
    _maxPer = str()
    _maxLa = float()
    _sep = " "
    _SPPList = {}
    _itemList = []
    _last = int()
    _finalPatterns = {}
    _tsList = {}
    _startTime = float()
    _endTime = float()
    _memoryUSS = float()
    _memoryRSS = float()
    _Database = []

    def __init__(self, inputFile, minSup, maxPer, maxLa, sep='\t'):
        self._iFile = inputFile
        self._minSup = minSup
        self._maxPer = maxPer
        self._maxLa = maxLa
        self._sep = sep

    def _creatingItemsets(self):
        self._Database = []
        if isinstance(self._iFile, _ab._pd.DataFrame):
            if self._iFile.empty:
                print("its empty..")
            i = self._iFile.columns.values.tolist()
            if 'Transactions' in i:
                self._Database = self._iFile['Transactions'].tolist()
            if 'Patterns' in i:
                self._Database = self._iFile['Patterns'].tolist()
        if isinstance(self._iFile, str):
            if _ab._validators.url(self._iFile):
                data = _ab._urlopen(self._iFile)
                for line in data:
                    line.strip()
                    line = line.decode("utf-8")
                    temp = [i.rstrip() for i in line.split(self._sep)]
                    temp = [x for x in temp if x]
                    self._Database.append(temp)
            else:
                try:
                    with open(self._iFile, 'r', encoding='utf-8') as f:
                        for line in f:
                            line.strip()
                            temp = [i.rstrip() for i in line.split(self._sep)]
                            temp = [x for x in temp if x]
                            self._Database.append(temp)
                except IOError:
                    print("File Not Found")
                    quit()

    def _convert(self, value):
        """
        to convert the type of user specified minSup value
        :param value: user specified minSup value
        :return: converted type
        """
        if type(value) is int:
            value = int(value)
        if type(value) is float:
            value = (len(self._Database) * value)
        if type(value) is str:
            if '.' in value:
                value = float(value)
                value = (len(self._Database) * value)
            else:
                value = int(value)
        return value

    def _createSPPList(self):
        """
            to convert the the single length stable periodic patterns
        """
        tidLast = {}
        la = {}
        self._SPPList = {}
        self._tsList = {}
        for transaction in self._Database:
            ts = int(transaction[0])
            for item in transaction[1:]:
                if item not in self._SPPList:
                    la[item] = max(0, ts - self._maxPer)
                    self._SPPList[item] = [1, la[item]]
                    self._tsList[item] = [ts]
                else:
                    s = self._SPPList[item][0] + 1
                    la[item] = max(0, la[item] + ts - tidLast.get(item) - self._maxPer)
                    self._SPPList[item] = [s, max(la[item], self._SPPList[item][1])]
                    self._tsList[item].append(ts)
                tidLast[item] = ts
            self._last = ts
        for item in self._SPPList:
            la[item] = max(0, la[item] + self._last - tidLast[item] - self._maxPer)
            self._SPPList[item][1] = max(la[item], self._SPPList[item][1])
        self._SPPList = {k: v for k, v in self._SPPList.items() if v[0] >= self._minSup and v[1] <= self._maxLa}
        self._SPPList = {k: v for k, v in sorted(self._SPPList.items(), key=lambda x: x[1][0], reverse=True)}
        self._Generation(list(self._SPPList), set())

    def _Generation(self, GPPFList, CP):
        """
        To generate the patterns using depth-first search
        """
        for i in range(len(GPPFList)):
            item = GPPFList[i]
            CP1 = CP | {item}
            if CP != set():
                self._tsList['\t'.join(CP1)] = list(set(self._tsList['\t'.join(CP)]) & set(self._tsList[item]))
            la = self._calculateLa(self._tsList['\t'.join(CP1)])
            support = len(self._tsList['\t'.join(CP1)])
            if la <= self._maxLa and len(self._tsList['\t'.join(CP1)]) >= self._minSup:
                #CP = CP1
                self._finalPatterns['\t'.join(CP1)] = [support, la]
                if i+1 < len(GPPFList):
                    self._Generation(GPPFList[i+1:], CP1)

    def _calculateLa(self, tsList):
        """ To calculate the lability of a patterns based on its timestamps"""
        previous = 0
        la = 0
        tsList = sorted(tsList)
        laList = []
        for ts in tsList:
            la = max(0, la + ts - previous - self._maxPer)
            laList.append(la)
            previous = ts
            
        la = max(0, la + self._last - previous - self._maxPer)
        laList.append(la)
        maxla = max(laList)
        return maxla

    def startMine(self):
        """ Method to start the mining of patterns"""
        self._startTime = _ab._time.time()
        self._creatingItemsets()
        self._minSup = self._convert(self._minSup)
        self._maxPer = self._convert(self._maxPer)
        self._maxLa = self._convert(self._maxLa)
        self._finalPatterns = {}
        #print(self._minSup, self._maxPer, self._maxLa)
        self._createSPPList()
        self._endTime = _ab._time.time()
        self._memoryUSS = float()
        self._memoryRSS = float()
        process = _ab._psutil.Process(_ab._os.getpid())
        self._memoryUSS = process.memory_full_info().uss
        self._memoryRSS = process.memory_info().rss
        print("Stable Periodic Frequent patterns were generated successfully using SPPECLAT algorithm ")


    def getRuntime(self):
        """Calculating the total amount of runtime taken by the mining process

        :return: returning total amount of runtime taken by the mining process
        :rtype: float
        """
        return self._endTime - self._startTime

    def getPatterns(self):
        """ Function to return the set of stable periodic-frequent patterns after completion of the mining process

                :return: returning stable periodic-frequent patterns
                :rtype: dict
        """
        return self._finalPatterns

    def getMemoryUSS(self):
        """Total amount of USS memory consumed by the mining process will be retrieved from this function
        :return: returning USS memory consumed by the mining process
        :rtype: float
        """

        return self._memoryUSS

    def save(self, outFile):
        """Complete set of periodic-frequent patterns will be loaded in to a output file

        :param outFile: name of the output file
        :type outFile: file
        """
        self._oFile = outFile
        writer = open(self._oFile, 'w+')
        for x, y in self._finalPatterns.items():
            s1 = x.strip() + ":" + str(y[0]) + ":" + str(y[1])
            writer.write("%s \n" % s1)

    def getPatternsAsDataFrame(self):
        """Storing final periodic-frequent patterns in a dataframe

        :return: returning periodic-frequent patterns in a dataframe
        :rtype: pd.DataFrame
        """

        dataFrame = {}
        data = []
        for a, b in self._finalPatterns.items():
            data.append([a.replace('\t', ' '), b[0], b[1]])
            dataFrame = _ab._pd.DataFrame(data, columns=['Patterns', 'Support', 'Periodicity'])
        return dataFrame

    def getMemoryRSS(self):
        """Total amount of RSS memory consumed by the mining process will be retrieved from this function
        :return: returning RSS memory consumed by the mining process
        :rtype: float
        """

        return self._memoryRSS

    def printResults(self):
        print("Total number of Stable Periodic  Patterns:", len(self.getPatterns()))
        print("Total Memory in USS:", self.getMemoryUSS())
        print("Total Memory in RSS", self.getMemoryRSS())
        print("Total ExecutionTime in ms:", self.getRuntime())

if __name__ == '__main__':
    _ap = str()
    if len(_ab._sys.argv) == 6 or len(_ab._sys.argv) == 7:
        if len(_ab._sys.argv) == 7:
            _ap = SPPEclat(_ab._sys.argv[1], _ab._sys.argv[3], _ab._sys.argv[4], _ab._sys.argv[5], _ab._sys.argv[6])
        if len(_ab._sys.argv) == 6:
            _ap = SPPEclat(_ab._sys.argv[1], _ab._sys.argv[3], _ab._sys.argv[4], _ab._sys.argv[5])
        _ap.startMine()
        print("Total number of Patterns:", len(_ap.getPatterns()))
        _ap.save(_ab._sys.argv[2])
        print("Total Memory in USS:", _ap.getMemoryUSS())
        print("Total Memory in RSS", _ap.getMemoryRSS())
        print("Total ExecutionTime in ms:", _ap.getRuntime())
    else:
        print("Error! The number of input parameters do not match the total number of parameters provided")



import os
import csv
import time
import numpy as np
import pycuda.gpuarray as gpuarray
import pycuda.autoinit
import psutil


class cudaAprioriGCT:
    __time = 0
    __memRSS = 0
    __memUSS = 0
    __GPU_MEM = 0
    filePath = ""
    sep = ""
    minSup = 0
    Patterns = {}

    def __init__(self, filePath, sep, minSup):
        self.filePath = filePath
        self.sep = sep
        self.minSup = minSup
        self.__time = 0
        self.__memRSS = 0
        self.__memUSS = 0

    def read_data(self, data_path, sep):
        data = []
        if not os.path.isfile(data_path):
            raise ValueError('Invalid data path.' + data_path)
        with open(data_path, 'r') as f:
            file = csv.reader(f, delimiter=sep, quotechar='\r')
            lineNo = 1
            for row in file:
                data.append([str(item) for item in row if item != ''])
                lineNo += 1
        return data, lineNo

    def write_result(self, result, result_path):
        file = open(result_path, 'w')
        for itemset, support in result.items():
            file.write(str(itemset) + ' : ' + str(support) + '\n')
        file.close()

    def compute_vertical_bitvector_data(self, data):
        #---build item to idx mapping---#
        idx = 0
        item2idx = {}
        for transaction in data:
            for item in transaction:
                if not item in item2idx:
                    item2idx[item] = idx
                    idx += 1
        idx2item = {idx: str(int(item)) for item, idx in item2idx.items()}
        #---build vertical data---#
        vb_data = np.zeros((len(item2idx), len(data)), dtype=np.uint16)
        for trans_id, transaction in enumerate(data):
            for item in transaction:
                vb_data[item2idx[item], trans_id] = 1
        vb_data = gpuarray.to_gpu(vb_data.astype(np.uint16))
        return vb_data, idx2item

    def getTime(self):
        return self.__time

    def getMEMORYRSS(self):
        return self.__memRSS

    def getMEMORYUSS(self):
        return self.__memUSS

    def getGPUMemory(self):
        return self.__GPU_MEM

    def getPatterns(self):
        return self.Patterns

    def get_numberOfPatterns(self):
        return len(self.Patterns)

    def startMine(self):
        startTime = time.time()
        basePattern = {}
        final = {}

        data, lineNo = self.read_data(self.filePath, self.sep)
        vb_data, idx2item = self.compute_vertical_bitvector_data(data)
        if self.minSup < 1:
            self.minSup = int(lineNo * self.minSup)

        for i in range(len(vb_data)):
            if gpuarray.sum(vb_data[i]).get() >= self.minSup:
                basePattern[idx2item[i]] = [i]
                final[idx2item[i]] = gpuarray.sum(vb_data[i]).get()

        while len(basePattern) > 0:
            temp = {}
            keysList = list(basePattern.keys())
            valuesList = list(basePattern.values())
            for i in range(len(basePattern)-1):
                keyI = keysList[i].split(" ")
                keyI = [int(x) for x in keyI]

                for j in range(i+1, len(basePattern)):
                    keyJ = keysList[j].split(" ")
                    keyJ = [int(x) for x in keyJ]
                    values = set(valuesList[i])
                    for val in valuesList[j]:
                        values.add(val)
                    values = list(sorted(values))
                    totalArray = vb_data[values[0]]
                    for k in range(1, len(values)):
                        totalArray = totalArray.__mul__(vb_data[values[k]])
                    support = gpuarray.sum(totalArray).get()
                    if support >= self.minSup:
                        combinedKey = " ".join(
                            str(x) for x in sorted(set(keyI) | set(keyJ)))
                        temp[combinedKey] = values
                        final[str(combinedKey)] = support
            basePattern = temp

        self.__time = time.time() - startTime
        self.__memRSS = psutil.Process(os.getpid()).memory_info().rss
        self.__memUSS = psutil.Process(os.getpid()).memory_full_info().uss
        self.Patterns = final
        self.__GPU_MEM = vb_data.nbytes

    def printStats(self):
        print("Total number of Coverage Patterns:", len(self.getPatterns()))
        print("Total Memory in USS:", self.getMemoryUSS())
        print("Total Memory in RSS", self.getMemoryRSS())
        print("Total ExecutionTime in ms:",  self.getRuntime())


if __name__ == "__main__":
    filePath = "datasets\\transactional_T10I4D100K.csv"
    sep = "\t"
    support = 500
    cudaAprioriGCT = cudaAprioriGCT(filePath, sep, support)
    cudaAprioriGCT.startMine()
    print("Time: ", cudaAprioriGCT.get_time())
    print("Memory RSS: ", cudaAprioriGCT.get_memRSS())
    print("Memory USS: ", cudaAprioriGCT.get_memUSS())
    print("GPU MEM: ", cudaAprioriGCT.get_GPU_MEM())
    print("Patterns: ", cudaAprioriGCT.get_numberOfPatterns())



from PAMI.coveragePatterns.basic import abstract as _ab

class CMine(_ab._coveragePatterns):
    """
        CMine algorithms aims to discover the coverage patterns in transactional databases.

    Reference:
    ---------
        Bhargav Sripada, Polepalli Krishna Reddy, Rage Uday Kiran:
        Coverage patterns for efficient banner advertisement placement. WWW (Companion Volume) 2011: 131-132
        https://dl.acm.org/doi/10.1145/1963192.1963259
    
    Attributes:
    -----------
        self.iFile : str
            Input file name or path of the input file
        minSup: float or int or str
            The user can specify minSup either in count or proportion of database size.
            If the program detects the data type of minSup is integer, then it treats minSup is expressed in count.
            Otherwise, it will be treated as float.
            Example: minSup=10 will be treated as integer, while minSup=10.0 will be treated as float
        sep : str
            This variable is used to distinguish items from one another in a transaction. The default separator is tab space or \t.
            However, the users can override their default separator.
        self.oFile : str
            Name of the output file or path of the output file
        self.startTime:float
            To record the start time of the mining process
        self.endTime:float
            To record the completion time of the mining process
        self.finalPatterns: dict
            Storing the complete set of patterns in a dictionary variable
        self.memoryUSS : float
            To store the total amount of USS memory consumed by the program
        self.memoryRSS : float
            To store the total amount of RSS memory consumed by the program
        self.Database : list
            To store the complete set of transactions available in the input database/file
    Methods:
    -------
        startMine()
            Mining process will start from here
        getPatterns()
            Complete set of patterns will be retrieved with this function
        save(oFile)
            Complete set of coverage patterns will be loaded in to a output file
        getPatternsAsDataFrame()
            Complete set of coverage patterns will be loaded in to a dataframe
        getMemoryUSS()
            Total amount of USS memory consumed by the mining process will be retrieved from this function
        getMemoryRSS()
            Total amount of RSS memory consumed by the mining process will be retrieved from this function
        getRuntime()
            Total amount of runtime taken by the mining process will be retrieved from this function
        createCoverageItems()
            Generate coverage items
        tidToBitset(itemset)
            Convert tid list to bit set
        genPatterns(prefix, tidData)
            Generate coverage patterns
        generateAllPatterns(coverageItems)
            Generate all coverage patterns

    Executing the code on terminal:
    -------
        Format:
        ------
            python3 CMine.py <inputFile> <outputFile> <minRF> <minCS> <maxOR> <'\t'>

        Examples:
        --------
            python3 CMine.py sampleTDB.txt patterns.txt 0.4 0.7 0.5 ','

        Sample run of importing the code:
        -------------------

            from PAMI.coveragePattern.basic import CMine as alg

            obj = alg.CMine(iFile, minRF, minCS, maxOR, seperator)

            obj.startMine()

            coveragePatterns = obj.getPatterns()

            print("Total number of coverage Patterns:", len(coveragePatterns))

            obj.save(oFile)

            Df = obj.getPatternsAsDataFrame()

            memUSS = obj.getMemoryUSS()

            print("Total Memory in USS:", memUSS)

            memRSS = obj.getMemoryRSS()

            print("Total Memory in RSS", memRSS)

            run = obj.getRuntime()

            print("Total ExecutionTime in seconds:", run)

    Credits:
    -------
        The complete program was written by P.Likhitha  under the supervision of Professor Rage Uday Kiran.
    """

    _startTime = float()
    _endTime = float()
    _finalPatterns = {}
    _iFile = " "
    _oFile = " "
    _sep = " "
    _minCS = str()
    _minRF = str()
    _maxOR = str()
    _memoryUSS = float()
    _memoryRSS = float()
    _Database = []
    _mapSupport = {}
    _lno = 0


    def _convert(self, value):
        """
        To convert the user specified minSup value
        :param value: user specified minSup value
        :return: converted type
        """
        if type(value) is int:
            value = int(value)
        if type(value) is float:
            value = value
        if type(value) is str:
            if '.' in value:
                value = float(value)
            else:
                value = int(value)
        return value

    def _creatingItemSets(self):
        """
            Storing the complete transactions of the database/input file in a database variable
        """
        self._Database = []
        self._mapSupport = {}
        if isinstance(self._iFile, _ab._pd.DataFrame):
            if self._iFile.empty:
                print("its empty..")
            i = self._iFile.columns.values.tolist()
            if 'Transactions' in i:
                self._Database = self._iFile['Transactions'].tolist()

        if isinstance(self._iFile, str):
            if _ab._validators.url(self._iFile):
                data = _ab._urlopen(self._iFile)
                for line in data:
                    line.strip()
                    line = line.decode("utf-8")
                    temp = [i.rstrip() for i in line.split(self._sep)]
                    temp = [x for x in temp if x]
                    self._Database.append(temp)
            else:
                try:
                    with open(self._iFile, 'r') as f:
                        for line in f:
                            self._lno += 1
                            splitter = [i.rstrip() for i in line.split(self._sep)]
                            splitter = [x for x in splitter if x]
                            self._Database.append(splitter)
                except IOError:
                    print("File Not Found")

    def creatingCoverageItems(self):
        """
        This function creates coverage items from _database.
        :return: coverageTidData that stores coverage items and their tid list.
        """
        tidData = {}
        self._lno = 0
        for transaction in self._Database:
            self._lno = self._lno + 1
            for item in transaction[1:]:
                if item not in tidData:
                    tidData[item] = [self._lno]
                else:
                    tidData[item].append(self._lno)
        coverageTidData = {k: v for k, v in tidData.items() if len(v) / len(self._Database) >= self._minRF}
        coverageTidData = dict(sorted(coverageTidData.items(), reverse=True, key=lambda x: len(x[1])))
        return coverageTidData

    def tidToBitset(self,itemset):
        """
        This function converts tid list to bitset.
        :param itemset:
        :return:
        """
        bitset = {}

        for k,v in itemset.items():
            bitset[k] = 0b1
            bitset[k] = (bitset[k] << int(v[0])) | 0b1
            for i in range(1,len(v)):
                diff = int(v[i]) - int(v[i-1])
                bitset[k] = (bitset[k] << diff) | 0b1
            bitset[k] = (bitset[k] << (self._lno - int(v[i])))
        return bitset

    def genPatterns(self,prefix,tidData):
        """
        This function generate coverage pattern about prefix.
        :param prefix: String
        :param tidData: list
        :return:
        """
        # variables to store coverage item set and
        itemset = prefix[0]

        # Get the length of tidData
        length = len(tidData)
        for i in range(length):
            tid = prefix[1] & tidData[i][1]
            tid1 = prefix[1] | tidData[i][1]
            andCount = bin(tid).count("1") - 1
            orCount = bin(tid1).count("1") - 1
            if orCount/len(self._Database) >= self._minCS and andCount / len(str(prefix[1])) <= self._maxOR:
                coverageItemset = itemset + '\t' + tidData[i][0]
                if orCount / len(self._Database) >= self._minRF:
                    self._finalPatterns[coverageItemset] = andCount
                self.genPatterns((coverageItemset,tid),tidData[i+1:length])

    def generateAllPatterns(self,coverageItems):
        """
        This function generates all coverage patterns.
        :param coverageItems: coverage items
        :return:
        """
        tidData = list(coverageItems.items())
        length = len(tidData)
        for i in range(length):
            #print(i,tidData[i][0])
            self.genPatterns(tidData[i],tidData[i+1:length])

    def startMine(self):
        """ Main method to start """

        self._startTime = _ab._time.time()
        if self._iFile is None:
            raise Exception("Please enter the file path or file name:")
        self._creatingItemSets()
        self._minCS = self._convert(self._minCS)
        self._minRF =  self._convert(self._minRF)
        self._maxOR = self._convert(self._maxOR)
        coverageItems = self.creatingCoverageItems()
        self._finalPatterns = {k: len(v) for k, v in coverageItems.items()}
        coverageItemsBitset = self.tidToBitset(coverageItems)
        self.generateAllPatterns(coverageItemsBitset)
        self.save('output.txt')
        self._endTime = _ab._time.time()
        process = _ab._psutil.Process(_ab._os.getpid())
        self._memoryUSS = float()
        self._memoryRSS = float()
        self._memoryUSS = process.memory_full_info().uss
        self._memoryRSS = process.memory_info().rss
        print("Coverage patterns were generated successfully using CMine  algorithm")

    def getMemoryUSS(self):
        """Total amount of USS memory consumed by the mining process will be retrieved from this function
        :return: returning USS memory consumed by the mining process
        :rtype: float
        """

        return self._memoryUSS

    def getMemoryRSS(self):
        """Total amount of RSS memory consumed by the mining process will be retrieved from this function
        :return: returning RSS memory consumed by the mining process
        :rtype: float
        """

        return self._memoryRSS

    def getRuntime(self):
        """Calculating the total amount of runtime taken by the mining process
        :return: returning total amount of runtime taken by the mining process
        :rtype: float
        """

        return self._endTime - self._startTime

    def getPatternsAsDataFrame(self):
        """Storing final coverage patterns in a dataframe
        :return: returning coverage patterns in a dataframe
        :rtype: pd.DataFrame
        """

        dataFrame = {}
        data = []
        for a, b in self._finalPatterns.items():
            data.append([a.replace('\t', ' '), b])
            dataFrame = _ab._pd.DataFrame(data, columns=['Patterns', 'Support'])
        return dataFrame

    def save(self, outFile):
        """Complete set of coverage patterns will be loaded in to a output file
        :param outFile: name of the output file
        :type outFile: file
        """
        self._oFile = outFile
        writer = open(self._oFile, 'w+')
        for x, y in self._finalPatterns.items():
            patternsAndSupport = x.strip() + ":" + str(y)
            writer.write("%s \n" % patternsAndSupport)

    def getPatterns(self):
        """ Function to send the set of coverage patterns after completion of the mining process
        :return: returning coverage patterns
        :rtype: dict
        """
        return self._finalPatterns

    def printResults(self):
        print("Total number of Coverage Patterns:", len(self.getPatterns()))
        print("Total Memory in USS:", self.getMemoryUSS())
        print("Total Memory in RSS", self.getMemoryRSS())
        print("Total ExecutionTime in ms:",  self.getRuntime())


if __name__=="__main__":
    _ap = str()
    if len(_ab._sys.argv) == 7 or len(_ab._sys.argv) == 6:
        if len(_ab._sys.argv) == 7:
            _ap = CMine(_ab._sys.argv[1], _ab._sys.argv[3], _ab._sys.argv[4], _ab._sys.argv[5], _ab._sys.argv[6])
        if len(_ab._sys.argv) == 6:
            _ap = CMine(_ab._sys.argv[1], _ab._sys.argv[3], _ab._sys.argv[4], _ab._sys.argv[5])
        _ap.startMine()
        print("Total number of coverage Patterns:", len(_ap.getPatterns()))
        _ap.save(_ab._sys.argv[2])
        print("Total Memory in USS:", _ap.getMemoryUSS())
        print("Total Memory in RSS", _ap.getMemoryRSS())
        print("Total ExecutionTime in ms:", _ap.getRuntime())
    else:
        _ap = CPPG('sample.txt', 0.4, 0.7, 0.5, ' ')
        _ap.startMine()
        print("Total number of Coverage Patterns:", len(_ap.getPatterns()))
        _ap.save('output.txt')
        _ap.printResults()
        print("Error! The number of input parameters do not match the total number of parameters provided")
